From 4e6ba90908c8d95da18f9329c096d83fa4fbb297 Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Hugo=20Beauz=C3=A9e-Luyssen?= <beauze.h@gmail.com>
Date: Sun, 22 Jan 2012 20:31:05 +0100
Subject: [PATCH 11/11] bluray: Use refcounting to avoid accessing invalid
 fields from vout.

---
 modules/access/bluray.c |  148 ++++++++++++++++++++++++-----------------------
 1 files changed, 75 insertions(+), 73 deletions(-)

diff --git a/modules/access/bluray.c b/modules/access/bluray.c
index 874307b..ed85635 100644
--- a/modules/access/bluray.c
+++ b/modules/access/bluray.c
@@ -79,6 +79,8 @@ typedef enum    OverlayStatus
 
 typedef struct bluray_overlay_t
 {
+    VLC_GC_MEMBERS
+
     vlc_mutex_t             lock;
     subpicture_t            *p_pic;
     OverlayStatus           status;
@@ -95,7 +97,7 @@ struct  demux_sys_t
     input_title_t       **pp_title;
 
     /* Overlay stuff */
-    bluray_overlay_t    overlay[MAX_OVERLAY];
+    bluray_overlay_t    *p_overlays[MAX_OVERLAY];
     int                 current_overlay; // -1 if no current overlay;
 
     input_thread_t      *p_input;
@@ -156,8 +158,6 @@ static int blurayOpen( vlc_object_t *object )
     {
         return VLC_ENOMEM;
     }
-    for ( int i = 0; i < MAX_OVERLAY; i++ )
-        vlc_mutex_init( &p_sys->overlay[i].lock );
     p_sys->current_overlay = -1;
 
     /* init demux info fields */
@@ -301,8 +301,6 @@ static void blurayClose( vlc_object_t *object )
         vlc_object_release( p_sys->p_input );
     if (p_sys->p_parser)
         stream_Delete(p_sys->p_parser);
-    for ( int i = 0; i < MAX_OVERLAY; ++i )
-    vlc_mutex_destroy( &p_sys->overlay[i].lock );
     /* Titles */
     for (unsigned int i = 0; i < p_sys->i_title; i++)
         vlc_input_title_Delete(p_sys->pp_title[i]);
@@ -376,6 +374,8 @@ static void     subpictureUpdaterUpdate( subpicture_t *p_subpic,
      * We need to duplicate our regions (stored internaly) to this subpic.
      */
     vlc_mutex_lock( &p_overlay->lock );
+    if ( p_overlay->p_regions == NULL )
+        return ;
     p_subpic->p_region = subpicture_region_Clone( p_overlay->p_regions );
     subpicture_region_t *p_src = p_overlay->p_regions->p_next;
     subpicture_region_t *p_dst = p_subpic->p_region;
@@ -394,48 +394,39 @@ static void     subpictureUpdaterUpdate( subpicture_t *p_subpic,
 
 static void     subpictureUpdaterDestroy( subpicture_t *p_subpic )
 {
-    free( p_subpic->updater.p_sys );
+    vlc_gc_decref( p_subpic->updater.p_sys->p_overlay );
 }
 
-/*****************************************************************************
- * libbluray overlay callbacks:
- *****************************************************************************/
-
-static void blurayDeleteOverlaySubPic( demux_t *p_demux, int i_plane )
+static void     blurayCleanOverayStruct( gc_object_t *p_gc )
 {
-    demux_sys_t     *p_sys = p_demux->p_sys;
+    bluray_overlay_t *p_overlay = vlc_priv( p_gc, bluray_overlay_t );
 
-    vlc_mutex_lock( &p_sys->overlay[i_plane].lock );
-    p_sys->overlay[i_plane].status = Closed;
-    if ( p_sys->overlay[i_plane].p_pic == NULL )
-    {
-        vlc_mutex_unlock( &p_sys->overlay[i_plane].lock );
-        return ;
-    }
-    //If we have a vout, let it cleanup by itself.
-    if ( p_sys->p_vout )
-        vout_FlushSubpictureChannel( p_sys->p_vout, p_sys->overlay[i_plane].p_pic->i_channel );
-    else
-    {
-        /*
-         * Else, the subpicture hasn't been sent yet, clean it ourselves.
-         * In that case, the vout doesn't hold the picture, we can use it safely
-         */
-        subpicture_Delete( p_sys->overlay[i_plane].p_pic );
-    }
-    p_sys->overlay[i_plane].p_pic = NULL;
-    vlc_mutex_unlock( &p_sys->overlay[i_plane].lock );
+    //This will be called when destroying the picture. Don't delete it again from here!
+    vlc_mutex_destroy( &p_overlay->lock );
+    subpicture_region_Delete( p_overlay->p_regions );
+    free( p_overlay );
 }
 
-static void blurayCloseOverlay( demux_t *p_demux )
+/*****************************************************************************
+ * libbluray overlay callbacks:
+ *****************************************************************************/
+
+static void blurayCloseAllOverlays( demux_t *p_demux )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
-    for ( int i = 0; i < MAX_OVERLAY; ++i )
-        blurayDeleteOverlaySubPic( p_demux, i );
     p_demux->p_sys->current_overlay = -1;
     if ( p_sys->p_vout != NULL )
     {
+        for ( int i = 0; i < 0; i++ )
+        {
+            if ( p_sys->p_overlays[i] != NULL )
+            {
+                vout_FlushSubpictureChannel( p_sys->p_vout,
+                                             p_sys->p_overlays[i]->p_pic->i_channel );
+                vlc_gc_decref( p_sys->p_overlays[i] );
+            }
+        }
         var_DelCallback( p_sys->p_vout, "mouse-moved", &onMouseEvent, p_demux );
         var_DelCallback( p_sys->p_vout, "mouse-clicked", &onMouseEvent, p_demux );
         vlc_object_release( p_sys->p_vout );
@@ -451,13 +442,13 @@ static void blurayActivateOverlay( demux_t *p_demux, const BD_OVERLAY* const ov
      * If the overlay is already displayed, mark the picture as outdated.
      * We must NOT use vout_PutSubpicture if a picture is already displayed.
      */
-    vlc_mutex_lock( &p_sys->overlay[ov->plane].lock );
-    if ( ( p_sys->overlay[ov->plane].status == Displayed ||
-           p_sys->overlay[ov->plane].status == Outdated )
+    vlc_mutex_lock( &p_sys->p_overlays[ov->plane]->lock );
+    if ( ( p_sys->p_overlays[ov->plane]->status == Displayed ||
+           p_sys->p_overlays[ov->plane]->status == Outdated )
          && p_sys->p_vout )
     {
-        p_sys->overlay[ov->plane].status = Outdated;
-        vlc_mutex_unlock( &p_sys->overlay[ov->plane].lock );
+        p_sys->p_overlays[ov->plane]->status = Outdated;
+        vlc_mutex_unlock( &p_sys->p_overlays[ov->plane]->lock );
         return ;
     }
     /*
@@ -466,21 +457,31 @@ static void blurayActivateOverlay( demux_t *p_demux, const BD_OVERLAY* const ov
      * the overlay is computed
      */
     p_sys->current_overlay = ov->plane;
-    p_sys->overlay[ov->plane].status = ToDisplay;
-    vlc_mutex_unlock( &p_sys->overlay[ov->plane].lock );
+    p_sys->p_overlays[ov->plane]->status = ToDisplay;
+    vlc_mutex_unlock( &p_sys->p_overlays[ov->plane]->lock );
 }
 
 static void blurayInitOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
-    if ( p_sys->overlay[ov->plane].p_pic != NULL )
-        blurayDeleteOverlaySubPic( p_demux, ov->plane );
+    assert( p_sys->p_overlays[ov->plane] == NULL );
 
+    p_sys->p_overlays[ov->plane] = calloc( 1, sizeof( **p_sys->p_overlays ) );
+    if ( unlikely( p_sys->p_overlays[ov->plane] == NULL ) )
+        return ;
     subpicture_updater_sys_t    *p_upd_sys = malloc( sizeof( *p_upd_sys ) );
     if ( unlikely( p_upd_sys == NULL ) )
+    {
+        free( p_sys->p_overlays[ov->plane] );
+        p_sys->p_overlays[ov->plane] = NULL;
         return ;
-    p_upd_sys->p_overlay = &p_sys->overlay[ov->plane];
+    }
+    vlc_gc_init( p_sys->p_overlays[ov->plane], blurayCleanOverayStruct );
+    //Incrementing refcounter: vout + demux
+    vlc_gc_incref( p_sys->p_overlays[ov->plane] );
+
+    p_upd_sys->p_overlay = p_sys->p_overlays[ov->plane];
     subpicture_updater_t    updater =
     {
         .pf_validate = subpictureUpdaterValidate,
@@ -488,22 +489,23 @@ static void blurayInitOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
         .pf_destroy = subpictureUpdaterDestroy,
         .p_sys = p_upd_sys,
     };
-    p_sys->overlay[ov->plane].p_pic = subpicture_New( &updater );
-    p_sys->overlay[ov->plane].p_pic->i_original_picture_width = ov->w;
-    p_sys->overlay[ov->plane].p_pic->i_original_picture_height = ov->h;
-    p_sys->overlay[ov->plane].p_pic->b_ephemer = true;
-    p_sys->overlay[ov->plane].p_pic->b_absolute = true;
+    p_sys->p_overlays[ov->plane]->p_pic = subpicture_New( &updater );
+    p_sys->p_overlays[ov->plane]->p_pic->i_original_picture_width = ov->w;
+    p_sys->p_overlays[ov->plane]->p_pic->i_original_picture_height = ov->h;
+    p_sys->p_overlays[ov->plane]->p_pic->b_ephemer = true;
+    p_sys->p_overlays[ov->plane]->p_pic->b_absolute = true;
 }
 
 static void blurayClearOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
-    vlc_mutex_lock( &p_sys->overlay[ov->plane].lock );
-    subpicture_region_ChainDelete( p_sys->overlay[ov->plane].p_regions );
-    p_sys->overlay[ov->plane].p_regions = NULL;
-    p_sys->overlay[ov->plane].status = Outdated;
-    vlc_mutex_unlock( &p_sys->overlay[ov->plane].lock );
+    vlc_mutex_lock( &p_sys->p_overlays[ov->plane]->lock );
+
+    subpicture_region_ChainDelete( p_sys->p_overlays[ov->plane]->p_regions );
+    p_sys->p_overlays[ov->plane]->p_regions = NULL;
+    p_sys->p_overlays[ov->plane]->status = Outdated;
+    vlc_mutex_unlock( &p_sys->p_overlays[ov->plane]->lock );
 }
 
 static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
@@ -516,7 +518,7 @@ static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
      */
     if ( ov->img )
     {
-        vlc_mutex_lock( &p_sys->overlay[ov->plane].lock );
+        vlc_mutex_lock( &p_sys->p_overlays[ov->plane]->lock );
 
         video_format_t          fmt;
         video_format_Init( &fmt, 0 );
@@ -551,11 +553,11 @@ static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
         }
         //Append the new region at the end of our current regions:
         //If there's no region yet, don't bother checking for update...
-        subpicture_region_t *p_reg = p_sys->overlay[ov->plane].p_regions;
+        subpicture_region_t *p_reg = p_sys->p_overlays[ov->plane]->p_regions;
         if ( p_reg == NULL )
         {
-            p_sys->overlay[ov->plane].p_regions = region;
-            vlc_mutex_unlock( &p_sys->overlay[ov->plane].lock );
+            p_sys->p_overlays[ov->plane]->p_regions = region;
+            vlc_mutex_unlock( &p_sys->p_overlays[ov->plane]->lock );
             return ;
         }
         /*
@@ -571,9 +573,9 @@ static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
             {
                 if ( p_prev == NULL )
                 {
-                    p_sys->overlay[ov->plane].p_regions = p_reg->p_next;
+                    p_sys->p_overlays[ov->plane]->p_regions = p_reg->p_next;
                     subpicture_region_Delete( p_reg );
-                    p_reg = p_sys->overlay[ov->plane].p_regions;
+                    p_reg = p_sys->p_overlays[ov->plane]->p_regions;
                 }
                 else
                 {
@@ -590,10 +592,10 @@ static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
         }
         /* If there's only one region, both prev and reg will be NULL */
         if ( p_prev == NULL )
-            p_sys->overlay[ov->plane].p_regions = region;
+            p_sys->p_overlays[ov->plane]->p_regions = region;
         else
             p_prev->p_next = region;
-        vlc_mutex_unlock( &p_sys->overlay[ov->plane].lock );
+        vlc_mutex_unlock( &p_sys->p_overlays[ov->plane]->lock );
     }
     /*
      * /!\ The region is now stored in our internal list, but not in the subpicture /!\
@@ -607,7 +609,7 @@ static void blurayOverlayProc( void *ptr, const BD_OVERLAY *const overlay )
     if ( overlay == NULL )
     {
         msg_Info( p_demux, "Closing overlay." );
-        blurayCloseOverlay( p_demux );
+        blurayCloseAllOverlays( p_demux );
         return ;
     }
     switch ( overlay->cmd )
@@ -664,22 +666,22 @@ static void bluraySendOverlayToVout( demux_t *p_demux )
     demux_sys_t *p_sys = p_demux->p_sys;
 
     //Sending subpicture to vout:
-    p_sys->overlay[p_sys->current_overlay].p_pic->i_start =
-            p_sys->overlay[p_sys->current_overlay].p_pic->i_stop = mdate();
-    p_sys->overlay[p_sys->current_overlay].p_pic->i_channel =
+    p_sys->p_overlays[p_sys->current_overlay]->p_pic->i_start =
+            p_sys->p_overlays[p_sys->current_overlay]->p_pic->i_stop = mdate();
+    p_sys->p_overlays[p_sys->current_overlay]->p_pic->i_channel =
             vout_RegisterSubpictureChannel( p_sys->p_vout );
     /*
      * Sending the subpicture to vout.
      * This must be done only once per subpicture, ie. only once between each
      * blurayInitOverlay & blurayCloseOverlay call.
      */
-    vout_PutSubpicture( p_sys->p_vout, p_sys->overlay[p_sys->current_overlay].p_pic );
+    vout_PutSubpicture( p_sys->p_vout, p_sys->p_overlays[p_sys->current_overlay]->p_pic );
 
     /*
      * Mark the picture as updated, as it contains no region for now.
      * This will make the subpicture_updater_t call pf_update
      */
-    p_sys->overlay[p_sys->current_overlay].status = Outdated;
+    p_sys->p_overlays[p_sys->current_overlay]->status = Outdated;
     //Activating vout events handling:
     var_AddCallback( p_sys->p_vout, "mouse-moved", &onMouseEvent, p_demux );
     var_AddCallback( p_sys->p_vout, "mouse-clicked", &onMouseEvent, p_demux );
@@ -998,17 +1000,17 @@ static int blurayDemuxMenu( demux_t *p_demux )
 
     if ( p_sys->current_overlay != -1 )
     {
-        vlc_mutex_lock( &p_sys->overlay[p_sys->current_overlay].lock );
-        if ( p_sys->overlay[p_sys->current_overlay].status == ToDisplay )
+        vlc_mutex_lock( &p_sys->p_overlays[p_sys->current_overlay]->lock );
+        if ( p_sys->p_overlays[p_sys->current_overlay]->status == ToDisplay )
         {
-            vlc_mutex_unlock( &p_sys->overlay[p_sys->current_overlay].lock );
+            vlc_mutex_unlock( &p_sys->p_overlays[p_sys->current_overlay]->lock );
             if ( p_sys->p_vout == NULL )
                 p_sys->p_vout = input_GetVout( p_sys->p_input );
             if ( p_sys->p_vout != NULL )
                 bluraySendOverlayToVout( p_demux );
         }
         else
-            vlc_mutex_unlock( &p_sys->overlay[p_sys->current_overlay].lock );
+            vlc_mutex_unlock( &p_sys->p_overlays[p_sys->current_overlay]->lock );
     }
     stream_DemuxSend( p_sys->p_parser, p_block );
     return 1;
-- 
1.7.8.4

